/*
	Copyright Andrew Varner 2011.07.22

	This file is part of rohttpfs.

	rohttpfs is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 3 as
	published by the Free Software Foundation.

	rohttpfs is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with rohttpfs.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "main.h"
#include <string>
#include <time.h>
#include <errno.h>
#include "file.h"
#include "rohttpfs.h"
#include "string.h"


file::file(const std::string path1, rohttpfs * server1)
{
	path = path1;
	server = server1;
	open_count = 0;
	size = 0;
	size_unknown = true;
	refresh_data();
	time_t now = time(NULL);
	age = now;
	idle = now;
}
file::~file()
{
	if (open_count != 0)
	{
		fprintf(logfile, "rohttpfs: warning: dtoring open file: %s\n", path.c_str());
	}
	while(data_cache.size() > 0)
	{
		file_block_element & i = data_cache.back();
		if (i.data != NULL)
		{
			delete[] i.data;
			server->remove_block(&i);
		}
		data_cache.pop_back();
	}
}

void file::open()
{
	time_t now = time(NULL);
	idle = now;
	open_count++;
	if (((age - now) > MAX_AGE_OF_FILE) && (type == 2))
		refresh_data();
	else if ((age - now) > MAX_AGE_OF_DIR)
		refresh_data();
}
void file::close()
{
	open_count--;
}
int file::read(char * buf, size_t requested_len, off_t offset)
{
	fprintf(logfile, "file::read f=\"%s\", r=%lu+%lu\n", path.c_str(), offset, requested_len);
	server->renew(path);
	switch (get_type())
	{
		case 0:
			return -ENOENT;
		case 1:
		case 2:
			break;
		case 3:
			return -EACCES;
		case 4:
			return -EACCES;
		default:
			return -EIO;
	}
	if (offset < 0)
		return -EIO;
	if (((size_t)offset) >= size)
		return 0;
	if (requested_len > (size - offset))
		requested_len = size - offset;
	
	if (size_unknown)
	{
		ssize_t read_len = read_direct(buf, requested_len, offset);
		if ((read_len >= 0) || (offset == 0))
			return read_len;
		else
		{
			char * temp = new char[offset + requested_len];
			if (temp == NULL)
			{
				fprintf(logfile, "out of memory :(\n");
				return -EIO;
			}
			memset(temp + offset, 0, requested_len);
			ssize_t num_read = read_direct(temp, offset + requested_len, 0);
			memcpy(buf, temp + offset, requested_len);
			delete[] temp;
			if (num_read <= 0)
				return num_read;
			else if (num_read < offset)
				return 0;
			else
				return num_read - offset;
		}
	}
	else
	{
		size_t block_num;
		size_t num_read = 0;
		while (num_read < requested_len)
		{
			block_num = (offset + num_read) / BLOCK_SIZE;
			size_t offset_block = (offset + num_read) % BLOCK_SIZE;
			size_t copy_size = BLOCK_SIZE - offset_block;
			if ((copy_size + num_read) > requested_len)
				copy_size = (requested_len - num_read);
			int status = fill_buffer(block_num, false);
			if (status != 0)
			{
				fprintf(logfile, "unable to get buffer\n");
				refresh_data();
				return status;
			}
			//atomic begin
			char * src = data_cache.at(block_num).data;
			if (src == NULL)
			{
				fprintf(logfile, "hard cache miss error, buy more memory\n");
				//this means that we're short on memory, and it was removed from cache right after it was put in.
				//atomic end
				return -EIO;
			}
			memcpy(buf + num_read, src + offset_block, copy_size);
			//atomic end
			num_read += copy_size;
		}

		//TODO: have a new thread speculatively call "fill_buffer(block_num + 1, false)"
	
		return num_read;
	}
}

ssize_t file::read_direct(char * buf, size_t requested_len, off_t offset)
{
	std::string complete_path = path;
	if (get_type() == 1)
		complete_path.append("/");
	
	int s = -1;
	ssize_t remote_len = 0;
	int status = server->send_request(&s, 0, NULL, 0, complete_path.c_str(), &remote_len, offset, requested_len + offset - 1);
	ssize_t retval = 0;
	if ((requested_len != (size_t)remote_len) && (remote_len >= 0))
		retval = -EIO;
	if (status == 404)
		retval = -ENOENT;
	if (status == 403)
		retval = -EACCES;
	if ((status != 206) && (status != 200))
		retval = -EIO;
	if ((offset != 0) && (status != 206))
		retval = -EIO;
	if (retval != 0)
	{
		if (s != -1)
			::close(s);
		return retval;
	}
	
	for (size_t filled = 0; filled < requested_len;  )
	{
		errno = 0;
		ssize_t read_len = ::read(s, buf + filled, requested_len - filled);
		if (errno != 0)
		perror("read errno");
		fprintf(logfile, "reading \"%s\", %ld+%lu/%lu\n", path.c_str(), read_len, filled, requested_len);
		/*
		if ((read_len < 0) && (errno == ECONNRESET) && (filled > 0))
		{
			::close(s);
			return filled;
		}
		*/
		if (read_len < 0)
		{
			perror("rohttpfs: socket read error");
			::close(s);
			return -EIO;
		}
		if (read_len == 0)
		{
			::close(s);
			return filled;
		}
		filled += read_len;
	}
	::close(s);
	return requested_len;
}

int file::fill_buffer(size_t block_num, bool force)
{
	fprintf(logfile, "filling buffer f=\"%s\" n=%lu\n", path.c_str(), block_num);
	size_t offset = block_num * BLOCK_SIZE;
	size_t request_size = BLOCK_SIZE;
	if (offset >= size)
	{
		fprintf(logfile, "error: request for buffer beyond maximum size of file. crash imminent\n");
		return -EIO;
	}
	if ((offset + request_size) >= size)
		request_size = size - offset;
	
	file_block_element & el = data_cache.at(block_num);
	if ((el.data != NULL) && (!force))
	{
		fprintf(logfile, "buffer is already cached\n");
		return 0;
	}

	char * buf = new char[request_size];
	if (buf == NULL)
	{
		fprintf(logfile, "out of memory\n");
		return -EIO;
	}
	ssize_t read_len = read_direct(buf, request_size, offset);
	if (((size_t)read_len) != request_size)
	{
		fprintf(logfile, "unable to fill entire buffer, only %ld\n", read_len);
		delete[] buf;
		if (read_len < 0)
			return read_len;
		else
			return -EIO;
	}
	char * deleteme = NULL;
	//atomic start
	file_block_element & el2 = data_cache.at(block_num);
	if (el2.data != NULL)
		deleteme = el2.data;
	else
		server->put_new_file_block(&el);
	el2.data = buf;
	//atomic end
	fprintf(logfile, "buffer filled\n");
	if (deleteme != NULL)
		delete[] deleteme;
	return 0;
}

std::string file::get_path()
{
	return path;
}
int file::get_type()
{
	return type;
}
size_t file::get_size()
{
	return size;
}
time_t file::get_mtime()
{
	return mtime;
}
void file::refresh_data()
{
	size_t size2;
	time_t mtime2;
	int type2 = server->get_file_info(path.c_str(), &size2, &mtime2);
	fprintf(logfile, "refreshing \"%s\": s=%ld; type=%d; mtime=%ld\n", path.c_str(), size2, type2, mtime2);
	if ((type != type2) || (size != size2) || (mtime != mtime2))
	{
		type = type2;
		size = size2;
		mtime = mtime2;
		size_unknown = (size == SIZE_UNDEFINED);
		
		while(data_cache.size() > 0)
		{
			file_block_element & i = data_cache.back();
			if (i.data != NULL)
			{
				delete[] i.data;
				server->remove_block(&i);
			}
			data_cache.pop_back();
		}
		file_block_element element_default;
		element_default.data = NULL;
		size_t num_blocks = (size + BLOCK_SIZE - 1) / BLOCK_SIZE;
		if (size_unknown)
			num_blocks = 1;
		std::vector<file_block_element> data_cache_new(num_blocks, element_default);
		data_cache.swap(data_cache_new);
	}
	//TODO: fill in any other needed metadata here
}

uint64_t file::get_open_count()
{
	return open_count;
}

