/*
	Copyright Andrew Varner 2011.07.22

	This file is part of rohttpfs.

	rohttpfs is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 3 as
	published by the Free Software Foundation.

	rohttpfs is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with rohttpfs.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "main.h"
#include "rohttpfs.h"
#include <stdexcept>
#include <string.h>


file * rohttpfs::open(std::string name)
{
	fprintf(logfile, "rohttpfs debug info: fb=%lu\n", file_blocks.size());
	//TODO: this function should be entirely atomic
	std::map<std::string, file_element *>::iterator iter = map_by_name.find(name);
	file_element * f;
	if (iter == map_by_name.end())
	{
		fprintf(logfile, "adding file to cache: %s\n", name.c_str());
		if (list_by_age.size() >= NUM_OPEN_FILES_MAX)
		{
			file_element * oldest = list_by_age.front();
			if (oldest->f->get_open_count() > 0)
				return NULL; //TODO: note elsewhere that this signals out of memory
			list_by_age.erase(oldest->iter_list);
			map_by_name.erase(oldest->iter_map);
			delete oldest->f;
			delete oldest;
		}
		f = new file_element;
		f->f = new file(name, this);
		f->iter_map = map_by_name.insert(std::pair<std::string, file_element *>(name, f)).first;
		list_by_age.push_back(f);
		f->iter_list = (--list_by_age.end());
	}
	else
	{
		f = iter->second;
		renew(f);
	}
	f->f->open();
	return f->f;
}
void rohttpfs::renew(file_element * f)
{
	list_by_age.erase(f->iter_list);
	list_by_age.push_back(f);
	f->iter_list = (--list_by_age.end());
}
void rohttpfs::renew(std::string name)
{
	renew(&*map_by_name[name]);
}

bool rohttpfs::ping_server()
{
	if (send_request(NULL, 1, NULL, 0, "/", NULL, 0, -1) != 200)
		return false;
	//this next request needs to be on a file, not an automatic index (otherwise server might not properly accept byte ranges)
	//if (send_request(NULL, 0, NULL, 0, "/", NULL, 0, 1) != 206)
	//	return false;
	//TODO: make sure server sends Content-Length in response
	return true;
}


rohttpfs::rohttpfs(std::string remote_host_name1)
{
	remote_host_name = remote_host_name1; //TODO: use this to lookup the actual address
	memset(&remote_host, 0, sizeof(remote_host));
	remote_host.sin_family = AF_INET;
	remote_host.sin_port = htons(80);
	remote_host.sin_addr.s_addr = htonl(3232235600); //192.168.0.80
	remote_host.sin_addr.s_addr = htonl(3232235524); //192.168.0.4
	if (!ping_server())
		throw std::runtime_error(std::string("remote host is fucked"));
}

int rohttpfs::send_request(int * s, bool head_only, char * redir_location, size_t redir_location_length, const char * relative_path, int64_t * output_size, size_t range_1, size_t range_2)
{
	if (relative_path[0] != '/')
		fprintf(logfile, "warning: requested path does not start with slash: %s\n", relative_path);
	else
		relative_path++;
	//TODO: relative_path must be URL-encoded
	const char * method = head_only?"HEAD":"GET";
	char range[200];
	range[0] = 0;
	if ((range_1 != 0) || (range_2 != ((size_t)-1)))
		snprintf(range, sizeof(range) / sizeof(range[0]), "Range: bytes=%lu-%lu\n", range_1, range_2);
	char request[1000];
	request[0] = 0;
	int req_size = snprintf(request, sizeof(request) / sizeof(request[0]), "%s /rohttpfs/%s HTTP/1.0\nHost: %s\nConnection: close\n%s\n", method, relative_path, remote_host_name.c_str(), range);
	if (req_size >= (int)(sizeof(request) / sizeof(request[0])))
		return 0;
//fprintf(logfile, "%s", request);
	
	int sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0)
		return 0;
	if (connect(sock, (struct sockaddr *)&remote_host, sizeof(remote_host)) != 0)
	{
		close(sock);
		return 0;
	}
	if (write(sock, request, req_size) != req_size)
	{
		close(sock);
		return 0;
	}
	char rbuf[1000];
	int i;
	for (i = 0; i < 1000; i++)
	{
		int rlen = read(sock, rbuf + i, 1);
		if (rlen < 0)
		{
			close(sock);
			return 0;
		}
		if (rlen == 0)
			break;
		if (i >= 1)
			if ((rbuf[i] == '\n') && (rbuf[i - 1] == '\n'))
				break;
		if (rbuf[i] == '\r')
			i--;
	}
	if (i >= 1000)
	{
		close(sock);
		return 0;
	}
	rbuf[i] = 0;
	int retval;
	int unused1, unused2;
	int status = sscanf(rbuf, "HTTP/%d.%d %d ", &unused1, &unused2, &retval);
	if (status != 3)
	{
		close(sock);
		return 0;
	}

	if (output_size != NULL)
	{
		int64_t calculated_size = -1;
		const char * contentlengthptr = strstr(rbuf, "\nContent-Length: ");
		if (contentlengthptr != NULL)
			sscanf(contentlengthptr + 1, "Content-Length: %ld\n", &calculated_size);
		*output_size = calculated_size;
	}

	if (redir_location_length > 0)
	{
		if ((retval >= 300) && (retval <= 399))
		{
			const char * redir_location_ptr = strstr(rbuf, "\nLocation:");
			if (redir_location_ptr != NULL)
			{
				const char * redir_location_start_ptr = strchr(redir_location_ptr, ':') + 1;
				while ((*redir_location_start_ptr < ' ') && (*redir_location_start_ptr != '\n') && (*redir_location_start_ptr != 0))
					redir_location_start_ptr++;
				const char * redir_location_end_ptr = strchrnul(redir_location_start_ptr, '\n');
				size_t redir_location_len_new = redir_location_end_ptr - redir_location_start_ptr;
				if (redir_location_len_new > redir_location_length)
				{
					strncpy(redir_location, redir_location_start_ptr, redir_location_length);
					redir_location[redir_location_length - 1] = 0;
				}
				else
				{
					strncpy(redir_location, redir_location_start_ptr, redir_location_len_new);
					redir_location[redir_location_len_new] = 0;
				}
			}
		}
		else
		{
			redir_location[0] = 0;
		}
	}
	
	if (head_only)
	{
		close(sock);
		sock = -1;
	}
	if (s != NULL)
	{
		*s = sock;
	}
	else
	{
		if (sock != -1)
			close(sock);
	}
	return retval;
}


void rohttpfs::put_new_file_block(file_block_element * f)
{
	while (((file_blocks.size() + 1) * BLOCK_SIZE > MAX_MEMORY) && (file_blocks.size() > 0))
		remove_block(file_blocks.front());
	//TODO: atomic (starting here);
	file_blocks.push_back(f);
	f->iter_list = (--file_blocks.end());
}


void rohttpfs::renew_block(file_block_element * f)
{
	//TODO: atomic
	file_blocks.erase(f->iter_list);
	file_blocks.push_back(f);
	f->iter_list = (--file_blocks.end());
}
void rohttpfs::remove_block(file_block_element * f)
{
	//TODO: atomic
	file_blocks.erase(f->iter_list);
	delete[] f->data;
}


//TODO: thread safety (in all files)

